@@ -8,7 +8,7 @@
  * memorandum "A Kernel Model for Precision Timekeeping" by Dave Mills,
  * revised 31 January 1996. That document updates RFC1589.
  *
- * Copyright (c) 1996, 1997, 1998 by Ulrich Windl
+ * Copyright (c) 1996 - 1999 by Ulrich Windl
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -24,7 +24,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * $Id: enable_pps.c,v 1.8 1998/12/30 14:23:56 windl Exp $ */
+ * $Id: enable_pps.c,v 1.13 1999/01/14 21:11:40 windl Exp $ */
 #include	<unistd.h>
 #include	<time.h>
 #include	<stdlib.h>
@@ -48,10 +48,14 @@
 				 */
 #endif
 
-#define OFFSET_TOLERANCE	5000
-#define STABIL_TOLERANCE	50
+#define OFFSET_TOLERANCE	50000
+#define STABIL_TOLERANCE	(10 << SHIFT_USEC)
+#define	PPS_INSANE		(STA_PPSERROR|STA_PPSWANDER|STA_PPSJITTER)
 
-static	const char	id[] = "$Id: enable_pps.c,v 1.8 1998/12/30 14:23:56 windl Exp $";
+#define	ALL_BITS_ON(v, b)		(((v) & (b)) == (b))
+#define	ALL_BITS_OFF(v, b)		(((v) & (b)) == 0)
+
+static	const char	id[] = "$Id: enable_pps.c,v 1.13 1999/01/14 21:11:40 windl Exp $";
 
 static	char	polarity	= 'p';	/* p[ositive] | n[egative] | 0) */
 
@@ -115,16 +119,16 @@ static	void	write_status_bits(FILE *fp, int bits)
 	for ( b = 0; b < 16; ++b )
 	{
 		if ( (bits & 1) != 0 )
-			fprintf(fp, "%s%s", sep, status_str[b]), sep = ", ";
+			fprintf(fp, "%s%s", sep, status_str[b]), sep = "+";
 		bits >>= 1;
 	}
 }
 
 /* print PPS status and update *evp */
-static	int	pps_status(int retval, int status, struct ppsclockev *evp)
+static	int	pps_status(const char *tag, const int show_event,
+			   int retval, int status, struct ppsclockev *evp)
 {
 	static	struct ppsclockev	last;
-	static	char const	tag[] = "PPS: ";
 	int	result = -1;
 	char const	*state_str = NULL;
 	char const	*sep	= tag;
@@ -135,7 +139,7 @@ static	int	pps_status(int retval, int status, struct ppsclockev *evp)
 		memset(evp, 0, sizeof(*evp));
 	}
 	
-	if ( evp->serial != last.serial )
+	if ( show_event && evp->serial != last.serial )
 	{
 		long	secs	= evp->tv.tv_sec - last.tv.tv_sec;
 		long	evs	= evp->serial - last.serial;
@@ -165,10 +169,10 @@ static	int	pps_status(int retval, int status, struct ppsclockev *evp)
 
 	if ( !(status & STA_PPSSIGNAL) )
 		fprintf(stderr, "%sno pulse", sep), sep = ", ";
-	status &= (STA_PPSJITTER|STA_PPSERROR|STA_PPSWANDER);
+	status &= (STA_PPSJITTER|STA_PPSERROR|STA_PPSWANDER|STA_CLOCKERR);
 	if ( status != 0 )
 	{
-		fprintf(stderr, "%s", sep), sep = ", ";
+		fprintf(stderr, "%serrors=", sep), sep = ", ";
 		write_status_bits(stderr, status);
 	}
 	if ( sep != tag )
@@ -177,12 +181,13 @@ static	int	pps_status(int retval, int status, struct ppsclockev *evp)
 }
 
 /* do adjtimex and check for errors */
-static	int	verbose_adjtimex(struct timex *txp, struct ppsclockev *evp)
+static	int	verbose_adjtimex(const char *tag, const int show_event,
+				 struct timex *txp, struct ppsclockev *evp)
 {
 	int	state = adjtimex(txp);
 	if ( state > 0 )
 	{
-		pps_status(state, txp->status, evp);
+		pps_status(tag, show_event, state, txp->status, evp);
 	}
 	else if ( state < 0 )
 	{
@@ -220,21 +225,36 @@ static	int	update_status(struct timex *txp, struct ppsclockev *evp,
 		fprintf(stderr, "]");
 	}
 	fprintf(stderr, "%s\n", sep[0] == ',' ? ")" : "");
-	return( verbose_adjtimex(txp, evp) );
+	return( verbose_adjtimex("update_status(): ", 0, txp, evp) );
 }
 
 /* reset kernel PLL */
 static	void	restart(struct timex *txp, struct ppsclockev *evp)
 {
-	txp->modes =
-		MOD_ESTERROR | MOD_MAXERROR | MOD_FREQUENCY | ADJ_TICKADJ;
-	txp->esterror = txp->maxerror = 123;	/* just some dummy value */
+	txp->modes = MOD_ESTERROR | MOD_MAXERROR;
+	/* indicate that we are trying to synchronize */
+	txp->esterror = txp->maxerror = MAXPHASE / 1000;
+	txp->modes |= MOD_FREQUENCY | ADJ_TICKADJ;
 	txp->freq = 0;
-	txp->tickadj = 500;			/* boost adjtime() times 100 */
-	verbose_adjtimex(txp, evp);
-	update_status(txp, evp, STA_FLL | STA_PPSFREQ | STA_PPSTIME |
-		      STA_UNSYNC,
-		      STA_FREQHOLD | STA_PLL);
+	/* boost adjtime() times 100 (HZ*500 == 0.05s per second) */
+	txp->tickadj = 500;			
+	txp->modes |= MOD_TIMECONST;
+	txp->constant = 1;
+	verbose_adjtimex("restart(): ", 0, txp, evp);
+	update_status(txp, evp, STA_FLL | STA_PPSFREQ | STA_PPSTIME,
+		      STA_FREQHOLD | STA_PLL | STA_UNSYNC);
+}
+
+/* use current time constant from PPS discipline if it is working properly */
+static	void	update_constant(struct timex *txp, struct ppsclockev *evp)
+{
+	if ( (txp->status & STA_PPSSIGNAL) != 0 &&
+	     (txp->status & PPS_INSANE) == 0 )
+		txp->constant = txp->shift;		/* copy PPS PLL constant */
+	else
+		txp->constant = 1;
+	txp->modes = MOD_TIMECONST;
+	verbose_adjtimex("update_constant(): ", 0, txp, evp);
 }
 
 /* do offset adjustments. expects offset within bounds */
@@ -244,6 +264,7 @@ static	void	pll_adj_offset(struct timex *txp, struct ppsclockev *evp)
 	static	u_int  last_serial;
 	time_t	t = time(NULL);
 	long	offset = evp->tv.tv_usec;
+	long	tdelta, tconst;
 
 	if ( last_serial == evp->serial )
 		return;
@@ -254,18 +275,32 @@ static	void	pll_adj_offset(struct timex *txp, struct ppsclockev *evp)
 	txp->modes = MOD_MAXERROR | MOD_ESTERROR;
 	txp->esterror = txp->maxerror = offset >= 0 ? : -offset;
 	txp->offset = (txp->offset + offset) / 2;
-	if ( txp->offset > MAXPHASE / 2 && txp->offset < -MAXPHASE / 2 )
+	if ( txp->offset > MAXPHASE / 2 || txp->offset < -MAXPHASE / 2 )
 		txp->modes = ADJ_OFFSET_SINGLESHOT;
 	else
 		txp->modes |= MOD_OFFSET;
-	if ( ((txp->status & STA_FLL) != 0 && t - last_adj_time < MINSEC) ||
-	     ((txp->status & STA_PLL) != 0 &&
-	      t - last_adj_time < (1L << txp->constant)) )
+	tdelta = t - last_adj_time;
+	tconst = ALL_BITS_ON(txp->status, STA_PPSTIME|STA_PPSSIGNAL) ?
+		txp->shift : txp->constant;
+	if ( ((txp->status & (STA_FLL|STA_PPSFREQ)) != 0 &&
+	      (tdelta < MINSEC || tdelta < (1L << tconst))) ||
+	     ((txp->status & STA_PLL) != 0 && tdelta < (1L << tconst) / 2) )
 		return;			/* don't adjust too frequently */
 	else
 		last_adj_time = t;
-printf("pll_adj_offset: in=%ld, out=%ld\n", offset, txp->offset);
-	verbose_adjtimex(txp, evp);
+	printf("%s: modes=%04x, delta=%ld, const=%ld, in=%ld, out=%ld\n",
+	       "pll_adj_offset", txp->modes, tdelta, tconst, offset,
+	       txp->offset);
+	verbose_adjtimex("pll_adj_offset(): ", 0, txp, evp);
+}
+
+/* adjust offset if conditions are sane */
+static	void	update_offset(struct timex *txp, struct ppsclockev *evp)
+{
+	if ( ((txp->status & STA_PPSSIGNAL) != 0 &&
+	      (txp->status & PPS_INSANE) == 0) ||
+	     (txp->status & STA_PPSSIGNAL) == 0 )
+		pll_adj_offset(txp, evp);	/* uses pps_offset! */
 }
 
 /* print status message */
@@ -361,7 +396,7 @@ int	main(int argc, char *argv[])
 		polarity);
 	tx.modes = 0;
 	print_header();
-	verbose_adjtimex(&tx, &event);
+	verbose_adjtimex("init: ", 1, &tx, &event);
 	print_stats(&tx);
 	if ( !keep_settings )
 	{
@@ -371,8 +406,8 @@ int	main(int argc, char *argv[])
 	state = S_NO_PPS;
 	while ( monitor )
 	{
-		if ( (time_state = verbose_adjtimex(&tx, &event)) == 0 )
-			pps_status(time_state, tx.status, &event);
+		tx.modes = 0;
+		time_state = verbose_adjtimex("monitor: ", 1, &tx, &event);
 		print_stats(&tx);
 		switch ( state )
 		{
@@ -380,39 +415,34 @@ int	main(int argc, char *argv[])
 			update_status(&tx, &event,
 				      STA_PPSTIME | STA_PPSFREQ | STA_FLL,
 				      STA_PLL | STA_FREQHOLD);
-			if ( time_state != 0 )
-				break;
-			else if ( (tx.status & STA_PPSSIGNAL) != 0
-				  && (tx.status &
-				      (STA_PPSJITTER|STA_PPSERROR)) == 0
-				  && (tx.offset > -OFFSET_TOLERANCE
-				      && tx.offset < OFFSET_TOLERANCE) )
+			update_offset(&tx, &event);
+			if ( (tx.status & STA_PPSSIGNAL) != 0
+			     && (tx.status & PPS_INSANE) == 0
+			     && (tx.offset > -OFFSET_TOLERANCE
+				 && tx.offset < OFFSET_TOLERANCE) )
 			{
   				state = S_PPS_TIME;
 				status("S_NO_PPS: PPS detected and in range");
 			}
-			if ( (tx.status & STA_PPSSIGNAL) != 0 )
-			{
-				pll_adj_offset(&tx, &event);
-			}
 			break;
 		case S_PPS_TIME:	/* adjust the offset */
 			update_status(&tx, &event,
 				      STA_FLL | STA_PPSFREQ,
 				      STA_PPSTIME | STA_FREQHOLD | STA_PLL);
-			pll_adj_offset(&tx, &event);	/* uses pps_offset! */
-			if ( tx.stabil > (STABIL_TOLERANCE << SHIFT_USEC) ||
-			     (event.tv.tv_usec > 2 * OFFSET_TOLERANCE &&
+			update_offset(&tx, &event);
+			if ( (event.tv.tv_usec > 2 * OFFSET_TOLERANCE &&
 			      event.tv.tv_usec <
 			      1000000 - 2 * OFFSET_TOLERANCE) )
 			{
 				status("S_PPS_TIME: PPS got insane");
 				state = S_NO_PPS;
 			}
-			if ( time_state == 0 &&
-			     tx.offset > -OFFSET_TOLERANCE / 2 &&
-			     tx.offset < OFFSET_TOLERANCE / 2 &&
-			     tx.shift > PPS_SHIFT )
+			else if ( (tx.status & STA_PPSSIGNAL) != 0 &&
+				  (tx.status & PPS_INSANE) == 0 &&
+				  tx.offset > -OFFSET_TOLERANCE / 2 &&
+				  tx.offset <= OFFSET_TOLERANCE / 2 &&
+				  tx.stabil <= STABIL_TOLERANCE &&
+				  tx.shift > PPS_SHIFT )
 			{
   				state = S_PPS_FREQ;
 				status("S_PPS_TIME: PPS offset is low and stable");
@@ -422,15 +452,21 @@ int	main(int argc, char *argv[])
 			update_status(&tx, &event,
 				      STA_FLL | STA_FREQHOLD,
 				      STA_PPSTIME | STA_PPSFREQ | STA_PLL);
+			update_constant(&tx, &event);
+			update_offset(&tx, &event);
 			if ( tx.shift == PPS_SHIFT ||
-			     tx.offset > -OFFSET_TOLERANCE / 2 &&
-			     tx.offset < OFFSET_TOLERANCE / 2 &&
-			     !(tx.status & STA_PPSSIGNAL) )
+			     tx.offset < -OFFSET_TOLERANCE ||
+			     tx.offset >= OFFSET_TOLERANCE ||
+			     (tx.status & STA_PPSSIGNAL) == 0 ||
+			     (tx.status & ~STA_PPSJITTER & PPS_INSANE) != 0 )
 			{
 				state = S_NO_PPS;
-				status("S_PPS_FREQ: invalid PPS signal");
+				status("S_PPS_FREQ: insane PPS signal");
 			}
-			if ( time_state == 0 && tx.shift == PPS_SHIFTMAX )
+			else if ( tx.shift == PPS_SHIFTMAX &&
+				  tx.offset > -OFFSET_TOLERANCE / 4 &&
+				  tx.offset <= OFFSET_TOLERANCE / 4 &&
+				  tx.stabil <= STABIL_TOLERANCE / 2 )
 			{
 				status("PPS_FREQ: best quality PPS signal");
 				state = S_PPS_FLL;
@@ -441,7 +477,11 @@ int	main(int argc, char *argv[])
 				      STA_FREQHOLD,
 				      STA_PPSTIME | STA_PPSFREQ | STA_FLL |
 				      STA_PLL);
-			if ( tx.shift <= PPS_SHIFTMAX )
+			update_constant(&tx, &event);
+			if ( (tx.status & (STA_PPSSIGNAL)) == 0 ||
+			     (tx.status & ~STA_PPSJITTER & PPS_INSANE) != 0 ||
+			     tx.stabil > STABIL_TOLERANCE ||
+ 			     tx.shift < PPS_SHIFTMAX )
 			{
 				state = S_PPS_FREQ;
 				status("S_PPS_FLL: PPS signal lost quality");
